//
// layout.h
//
// Ulrik Petersen
//
// Based on code by Todd Hoatson and/or Fred Eichelman, 
// though heavily modified by Ulrik Petersen.
//
// The copyright for the original code was assigned to Ulrik Petersen
// by Todd Hoatson's superior, Steve Echerd, in a letter dated
// November 11, 2003.
//
// Created: 7/17-2005
// Last update: 3721-2006
//
//
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2005-2006  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation, license version 2.  
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *   02111-1307 USA
 *
 *
 *   Special exception
 *   =================
 * 
 *   In addition, as a special exception, Ulrik Petersen, the
 *   copyright holder of Emdros, gives permission to link Emdros, in
 *   whole or in part, with the libraries which are normally
 *   distributed with:
 *   
 *   a) Sun's Java platform,
 *   b) Python, 
 *   c) Jython,
 *   d) Ruby, and/or 
 *   e) Perl 
 *   f) PostgreSQL
 *   g) OpenSSL
 *
 *   (or with modified versions of these), and to distribute linked
 *   combinations including both Emdros, in whole or in part, and one
 *   or more of the libraries normally distributed with (a)-(g) above.
 *
 *   Please note: This gives you special rights concerning the
 *   libraries which normally accompany the above pieces of software.
 *   It gives you no special rights concerning software that you write
 *   yourself.  You must obey the GNU General Public License in all
 *   respects for all of the code used other than the libraries
 *   normally distributed with (a)-(g) above.
 *
 *   If you modify this file, you may extend this exception to your
 *   version of the file, but you are not obligated to do so. If you
 *   do not wish to do so, delete this exception statement from your
 *   version.
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 *
 *   Website
 *   =======
 *
 *   Emdros has a website here:
 *
 *   http://emdros.org
 *
 *
 *
 **************************************************************************/



#ifndef DISPBOX__H__
#define DISPBOX__H__

#if defined(__GNUG__) && !defined(__APPLE__)
#pragma interface "layout.cpp"
#endif



#include <wx/wx.h>
#include <wx/scrolwin.h>
#include <wx/string.h>
#include <wx/dc.h>
#include <wx/colour.h>
#include <vector>

#include <emdros-lconfig.h>
#if HAVE_OSTREAM
#include <ostream>
#else
#include <ostream.h>
#endif



typedef enum { BS_SMALLEST, BS_SMALL, BS_MEDIUM, BS_LARGE } Bracket_Size;
typedef enum { BD_LEFT, // stands at the left
	       BD_RIGHT   // stands at the right
} Bracket_Direction;

typedef enum { BT_BRACKET, BT_PARENTHESIS, BT_DOUBLE_BRACKET } Bracket_Type;



#include <list>

enum PLACEMENT { 
	FRONT, // At the (logical) beginning of the line 
	BACK   // At the (logical) end of the line
};



class ViewMetrics; // Forward declaration

typedef std::vector<wxString> wxstringvec_t;

class LayoutItem {
 protected:
	ViewMetrics *m_pMetrics;
	int m_xExtentLP;
	int m_yExtentLP;
 public:
	LayoutItem(ViewMetrics *pViewMetrics) 
		: m_pMetrics(pViewMetrics), 
		m_xExtentLP(0), 
		m_yExtentLP(0) {};
	virtual ~LayoutItem() {};
	int GetWidthLP(void) const { return m_xExtentLP; };
	int GetHeightLP(void) const { return m_yExtentLP; };
	virtual void toHTML(std::ostream *pOut) { /* By default, do nothing. */ };
};

class LayoutBox : public LayoutItem {
 protected:
	bool m_bIsSelected;
	bool m_bIsGrayed;
	wxstringvec_t m_strings;
	long m_long;
 public:
	LayoutBox(wxDC *pDC, const wxstringvec_t& strings, ViewMetrics *pMetrics);
	virtual ~LayoutBox() {};
	virtual void Draw(wxDC *pDC, int xLP, int yLP);
	virtual void DrawBox(wxDC *pDC, int xLeftLP, int yTopLP, int xRightLP, int yBottomLP);
	virtual void DrawText(wxDC *pDC, wxString strOutput, int xLP, int yLP, unsigned int ilmIndex = 0);
	bool IsSelected(void) const { return m_bIsSelected; };
	void SetSelected(bool selected) { m_bIsSelected = selected; };
	bool IsGrayed(void) const { return m_bIsGrayed; };
	void SetGrayed(bool grayed) { m_bIsGrayed = grayed; };
	long GetLong(void) const { return m_long; };
	void SetLong(long l) { m_long = l; };
	virtual void toHTML(std::ostream *pOut, bool bIsRightToLeft);
};

// One that has only 1 string-item,
// but which uses the first line of the viewmetrics items
// to get its font
class SimpleLayoutBox : public LayoutBox {
 public:
	SimpleLayoutBox(wxDC *pDC, const wxString& mystring, ViewMetrics *pMetrics);
	virtual ~SimpleLayoutBox() {};
};

// One that has only 1 string-item
// but which uses ViewMetric's latin font.
class LatinLayoutBox : public SimpleLayoutBox {
 protected:
	wxColour m_color;
 public:  
	LatinLayoutBox(wxDC *pDC, const wxString& mystring, ViewMetrics *pMetrics, wxColour color);
	~LatinLayoutBox();
	virtual void DrawText(wxDC *pDC, wxString strOutput, int xLP, int yLP, unsigned int ilmIndex = 0);
	virtual void toHTML(std::ostream *pOut, bool bIsRightToLeft);
};


////////////////////////////////////////////////////////////
//
// LayoutBracketBox
//
////////////////////////////////////////////////////////////

class LayoutBracketBox : public SimpleLayoutBox {
 private:
	Bracket_Size m_bracket_size;
	Bracket_Direction m_bracket_direction;
	Bracket_Type m_bracket_type;
	wxString m_label;
	bool m_has_arrow;
	bool m_has_label;
	bool m_bIsRightToLeft;
 public:
	LayoutBracketBox(wxDC *pDC, 
			 ViewMetrics *pVwMetrics, 
			 bool opening, bool is_inside, 
			 wxString label, 
			 Bracket_Size bs, 
			 Bracket_Type bt,
			 bool bIsRightToLeft);
	virtual ~LayoutBracketBox(void);
	virtual void Draw(wxDC *pDC, int x, int y);
	static bool should_have_label(bool opening, bool is_inside) { return !opening && !is_inside; };
	virtual void toHTML(std::ostream *pOut, bool bIsRightToLeft);
 private:
	int m_nSerifLength;
	int m_nLabelXExtent;
	int m_nHeight;
	int m_nInnerGap;
	int m_nOuterGap;
	int m_nLineWidth;
	int m_nArrowLength;
	int m_nArrowLineWidth;
	void calculate_parameters(ViewMetrics *pVwMetrics);
};

////////////////////////////////////////////////////////////
//
// LayoutLine
//
////////////////////////////////////////////////////////////

typedef std::list<LayoutBox*> PBoxList;




class LayoutLine : public LayoutItem { // LayoutItem is defined in DispBox.h
 protected:
	// Boxes in logical (not display) order
	PBoxList m_boxes;
	int m_indent;
	bool m_bIsRightToLeft;
	bool m_bIsCentered;
	bool m_bOwnsBoxes;
 public:
	LayoutLine(wxDC *pDC, ViewMetrics *pViewMetrics, int nNumIndents, bool bOwnsBoxes, bool bIsRightToLeft, bool bIsCentered = false);
	virtual ~LayoutLine();
	void AddBox(LayoutBox *pBox, PLACEMENT placement = BACK);
	LayoutBox *RemoveBox(PLACEMENT placement);
	void Draw(wxDC *pDC, int xLP, int yLP);
	LayoutBox *GetBoxForX(int targetX);
	virtual void toHTML(std::ostream *pOut);
};


////////////////////////////////////////////////////////////
//
// LayoutParagraph
//
////////////////////////////////////////////////////////////

typedef std::list<LayoutLine*> PLineList;


class LayoutParagraph : public LayoutItem {
protected:
	PLineList m_lines;
	PBoxList m_boxes;
	int m_xUsableWidthLP;
	bool m_bIsRightToLeft;
	bool m_bIsCentered;
	int m_indent;
public:
	LayoutParagraph(ViewMetrics *pViewMetrics, int xUsableWidthLP, bool bIsRightToLeft, bool bIsCentered = false, int indent = 0);
	virtual ~LayoutParagraph();
	void AddBox(wxDC *pDC, LayoutBox *pBox);
	void AddBox(wxDC *pDC, LayoutBox *pBox, bool bAddToM_Boxes, bool bRecalculateExtents = true);
	void RecalculateExtents(void);
	void Draw(wxDC *pDC, int xLP, int yLP, int first_line, int last_line);
	void rewrap(wxDC *pDC, int xUsableWidthLP);
	int GetHeightLP(int first_line, int last_line);
	int getNoOfLines(void) const { return (int) m_lines.size(); };
	LayoutLine *GetLineForY(int targetY);
	virtual void toHTML(std::ostream *pOut);
 protected:
	void delete_lines(void);
};


typedef std::list<LayoutParagraph*> PParagraphList;

class LayoutCanvas : public wxScrolledWindow {
 public:
	LayoutCanvas(wxWindow *pParent, const wxPoint& pos, const wxSize& size, const long style);
	virtual void OnDraw(wxDC &dc);
	void OnSize(wxSizeEvent& event);
	void OnMouseEvent(wxMouseEvent& event);
   
	DECLARE_EVENT_TABLE()
		public:
	int m_nMapMode;
	int m_nMagnification;
	ViewMetrics *m_pViewMetrics;
 public:
	virtual void ChangeMagnification(int magnification);
	int GetMagnification(void) const { return m_nMagnification; };
	ViewMetrics *GetViewMetrics(void) { return m_pViewMetrics; };
	virtual void InitializeViewMetrics(); // Call this AFTER initializing it!
	bool UpdateViewMetricsViewWidth(void);
	void AddParagraph(LayoutParagraph *pParagraph);
	void AddParagraph(bool bIsRightToLeft, bool bIsCentered = false, int indent = 0);
	void AddBox(LayoutBox *pBox);
	void AddBox(wxDC *pDC, LayoutBox *pBox, bool bRecalculateParagraphExtentsNow = true);
	LayoutBox *AddLayoutBox(const wxstringvec_t& strings, bool bIsSelected);
	LayoutBox *AddLayoutBox(wxDC *pDC, const wxstringvec_t& strings, bool bIsSelected, bool bRecalculateParagraphExtentsNow = true);
	void RecalculateParagraphExtents(void);
	void AddLatinLayoutBox(const wxString& str, wxColour color);
	void AddLayoutBracketBox(bool opening, bool is_inside, 
				 wxString label, 
				 Bracket_Size bs, 
				 Bracket_Type bt,
				 bool bIsRightToLeft,
				 bool bIsSelected);
	LayoutBox *GetBoxForPoint(wxPoint point);
	virtual void HandleButtonDown(wxMouseEvent& event, wxPoint point, bool RightButton) {};
	void flush(void);
	void Clear(void);
	virtual void toHTML(std::ostream *pOut);
 protected:	
	virtual ~LayoutCanvas();
	virtual void printDTDParagraphEntry(std::ostream *pOut, std::string name, wxString strFontName, wxColour color, int pointsize);
	virtual void printHTMLHeader(std::ostream *pOut);
	virtual void printHTMLFooter(std::ostream *pOut);
	void DrawItem(wxDC *pDC, LayoutParagraph *pPara, int y);
	void CalculateOverallHeight();
	void GetClipBox(wxDC *pDC, wxRect& clipRect);
	void rewrap_paragraphs(wxDC *pDC);
	PParagraphList m_paragraphs;
	double m_xScaleFactor, m_yScaleFactor;
	int m_iOverallHeight;  //height of all current paragraph items
};


#endif


