/*
 * mql_symboltable.h
 *
 * MQL symbol table for object references
 *
 * Ulrik Petersen
 * Created: 3/16-2001
 * Last update: 10/6-2006
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2006  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation, license version 2.  
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *   02111-1307 USA
 *
 *
 *   Special exception
 *   =================
 * 
 *   In addition, as a special exception, Ulrik Petersen, the
 *   copyright holder of Emdros, gives permission to link Emdros, in
 *   whole or in part, with the libraries which are normally
 *   distributed with:
 *   
 *   a) Sun's Java platform,
 *   b) Python, 
 *   c) Jython,
 *   d) Ruby, and/or 
 *   e) Perl 
 *   f) PostgreSQL
 *   g) OpenSSL
 *
 *   (or with modified versions of these), and to distribute linked
 *   combinations including both Emdros, in whole or in part, and one
 *   or more of the libraries normally distributed with (a)-(g) above.
 *
 *   Please note: This gives you special rights concerning the
 *   libraries which normally accompany the above pieces of software.
 *   It gives you no special rights concerning software that you write
 *   yourself.  You must obey the GNU General Public License in all
 *   respects for all of the code used other than the libraries
 *   normally distributed with (a)-(g) above.
 *
 *   If you modify this file, you may extend this exception to your
 *   version of the file, but you are not obligated to do so. If you
 *   do not wish to do so, delete this exception statement from your
 *   version.
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 *
 *   Website
 *   =======
 *
 *   Emdros has a website here:
 *
 *   http://emdros.org
 *
 *
 *
 **************************************************************************/

/**@file mql_symboltable.h
 *@brief Header file for MQL symbol table (MQL layer)
 */



#ifndef MQL_SYMBOLTABLE__H__
#define MQL_SYMBOLTABLE__H__

#include <emdf.h>
#include <infos.h>
#include <map>
#include <list>
#include <string_func.h>
#include <smart_vector.h>
#include <exception_emdros.h>
#include <emdfdb.h>
#include <functional>

/** MQL symbol table exception.
 */
class MQLSymbolTableException : public EmdrosException {
 public:
	/** Constructor: Give a default exception message. 
	 */
	MQLSymbolTableException() : EmdrosException("An MQL symbol table exception occurred.") {};

	/** Constructor: Set the exception message.
	 *
	 * @param message the message to use.
	 */
	MQLSymbolTableException(const std::string& message) : EmdrosException(message) {};

	/** Destructor. */
	~MQLSymbolTableException() {};
};

/** MQL symbol table entry exception.
 */
class MQLSymbolTableEntryException : public EmdrosException {
 public:
	/** Constructor: Give a default exception message. 
	 */
	MQLSymbolTableEntryException() : EmdrosException("An MQL symbol table exception occurred.") {};

	/** Constructor: Set the exception message.
	 *
	 * @param message the message to use.
	 */
	MQLSymbolTableEntryException(const std::string& message) : EmdrosException(message) {};

	/** Destructor. */
	~MQLSymbolTableEntryException() {};
};

class ObjectBlock; // Forward declaration
class MQLObject; // Forward declaration
class EMdFDB; // Forward declaration
class EMdFValue; // Forward declaration

/** A std::pair with <EMdFValue*,FeatureInfo> as the parameters.
 *@internal
 *
 * This is used in a ValPairVec, which is used in MQLSymbolTableEntry.
 */
typedef std::pair<EMdFValue*, FeatureInfo> ValPair;

/** predicate for saying whether the ValPair's behind two ValPair pointers
 * are equal.
 *@internal
 */
struct valpair_equal : public std::equal_to<ValPair*> {
	bool operator() (const ValPair* x, const ValPair* y) const { return strcmp_nocase(x->second.getName(),y->second.getName()) == 0; };
};

/** A SmarVector of ValPairs.
 *@internal
 */
typedef SmartVector<ValPair, valpair_equal> ValPairVec;

class MQLSymbolTableEntry {
 private:
	id_d_t m_object_type_id;
	std::string m_object_type_name;
	ValPairVec m_vec;
	std::string m_object_reference;
	id_d_t m_id_d;
	std::list<FeatureInfo> *m_pFeature_infos;
 public:
	MQLSymbolTableEntry(const std::string& object_reference,
			    const std::string& object_type_name,
			    id_d_t object_type_id);
	~MQLSymbolTableEntry();
	short int addFeature(const FeatureInfo& fi);
	void setFeature(short int index, EMdFValue *feature_value);
	const std::list<FeatureInfo>& getFeatureInfos(void);
	id_d_t getFeatureTypeId(short int index);
	void getFeatureValue(short int index, const EMdFValue **result);
	void setID_D(id_d_t id_d_O);
	id_d_t getID_D(void) const { return m_id_d; };
	const std::string& getObjectReference(void) const { return m_object_reference; };
	const std::string& getObjectTypeName(void) const { return m_object_type_name; };
	id_d_t getObjectTypeId(void) const { return m_object_type_id; };
	void setFeatures(EMdFDB* pDB) throw(EMdFDBDBError);
	void setFeatures(MQLObject *pMQLObject);
};

// StrSTEPairVec
/** A std::pair of std::string and MQLSymbolTableEntry*.
 *@internal
 */
typedef std::pair<std::string, MQLSymbolTableEntry*> StrSTEPair;
/** A binary predicate to compare StrSTEPair pointers for equality.
 *@internal
 */
struct strste_pair_equal : public std::equal_to<StrSTEPair*> {
	bool operator() (const StrSTEPair* x, const StrSTEPair* y) const { return strcmp_nocase(x->first,y->first) == 0; };
};
/** A SmartVector of StrSTEPair objects.
 *@internal
 */
typedef SmartVector<StrSTEPair, strste_pair_equal> StrSTEPairVec;

// StrCOBCPairVec
/** A std::pair of std::string and ObjectBlock*.
 *@internal
 */
typedef std::pair<std::string, ObjectBlock*> StrCOBCPair;
/** A binary predicate to compare StrCOBCPairs for equality.
 *@internal
 */
struct strcobc_pair_equal : public std::equal_to<StrCOBCPair*> {
	bool operator() (const StrCOBCPair* x, const StrCOBCPair* y) const { return strcmp_nocase(x->first,y->first) == 0; };
};
/** A smart vector of StrCOBCPair using strcobc_pair_equal to compare.
 *@internal
 */
typedef SmartVector<StrCOBCPair, strcobc_pair_equal> StrCOBCPairVec;

class MQLSymbolTable {
 private:
	StrSTEPairVec m_vec_ste;
	StrCOBCPairVec m_vec_object_block;
 public:
	MQLSymbolTable();
	~MQLSymbolTable();
	short int insert(MQLSymbolTableEntry* pSTE, ObjectBlock *pObjectBlock) throw(MQLSymbolTableException);
	bool exists(const std::string& object_reference, short int& index);
	MQLSymbolTableEntry* get(short int index);
	ObjectBlock* getObjectBlock(short int index);
	short int getIndex(const std::string& object_reference);
};


#endif /* MQL_SYMBOLTABLE__H__ */
